/** ----------------------------------- -o-
 * OSCSend.java  (mxj object)
 *
 *   Encode and send OSC commands intended for server or named client 
 *     at a particular IP + port.
 *   Modelled after Sciss server example.
 *
 *
 * OBJECT ARGUMENTS: <ip/addr> <port> (<clientID>|server) [<oscdomain>]
 *
 * OUTGOING OSC FORMAT FOR SERVER AND CLIENT (RESPECTIVELY):
 *   /<oscdomain>/C/clientID/<channel> [<command_token> [<arg(s)>]]
 *   /<oscdomain>/S/<request_token> [<arg(s)>]
 *
 * OUTGOING MAX MESSAGE FORMAT: <channel> [<command_token> [<arg(s)>]]
 *
 *
 * LIBRARY DEPENDENCIES: 
 *   NetUtil -- http://www.sciss.de/netutil/
 *   max.jar -- /Applications/Max5/Cycling '74/java/lib/max.jar
 *
 * BUILD AND INSTALL: 
 *   1) javac -classpath <cp> OSCSend.java
 *        <cp> = colon separated list of paths of jar file library dependencies.
 *   2) Place resulting class files in any directory under the root directory 
 *        containing the Max patch that uses "mxj OSCReceive.java".
 *      
 *
 * INLETS--
 *   0: Max messages (see below)
 *   1: OSC messages
 *
 * MESSAGES--   
 *   open / close
 *   oscdomain
 *
 * METHODS--  
 *   send
 *   setoscdomain
 * 
 * ----------------------------------------------------------------------
 * Copyright © 2011-2012, David Reeder.  All rights reserved.
 * Please see LICENSE file in this distribution for license details.
 * Contact: osc@davidreeder.info
 * ----------------------------------------------------------------------
 */ 	 								// -o--

import com.cycling74.max.*;
import de.sciss.net.*;

import java.net.InetAddress;
import java.net.SocketAddress;
import java.net.InetSocketAddress;
import java.io.IOException;



public class OSCSend extends MaxObject
{ 							
  //static private DebugLabel_M 	z = new DebugLabel_M("OSCSend", true);
  //static private ToolBox    	t = new ToolBox();

  int verbose = 1;


  private String 	USAGE = "Usage: OSCSend <IP/addr> <port> (<clientID>|server) [<oscdomain>]";
  private String 	version = "0.2";

  private String 	MXJERR = "MXJ argument error: ";


  private OSCServer 	server = null;

  private String  	send_ip;
  private int 		send_port;

  private InetAddress 	send_addr;
  private SocketAddress send_sockaddr;


  private enum TargetType 	{ CLIENT, SERVER };

  private String 	oscdomain = "domain"; 	// DEFAULT
  private String 	send_prefix = null;
  private TargetType 	send_targettype = null;
  private String 	send_targetId;




  //------------------------------------------------------------------- -o--
  // constructors
  //
  public OSCSend(Atom[] args) 
  { 
    //zsm("OSCSend");
    zsm("");

    zi("version " + version + ".  osc@davidreeder.info");

    if (2 <= verbose)  { zarglist(args); }


    //
    // send args and initialize
    //
    if ((3 > args.length) || (4 < args.length)) { 
      zerr(MXJERR + "Improper number of arguments.");
      zferr(USAGE); 
    }

    send_ip 	   = args[0].toString();
    send_port 	   = args[1].toInt();
    send_targetId  = args[2].toString();

    if (3 < args.length) { 
      oscdomain	   = args[3].toString();
    }


    if (send_targetId.equals("server")) {
      send_targettype = TargetType.SERVER;
    } else {
      send_targettype = TargetType.CLIENT;
    }

    setoscdomain(oscdomain);


    try {
      send_addr 	= InetAddress.getByName(send_ip);
      send_sockaddr 	= new InetSocketAddress(send_addr, send_port);

    } catch (Exception e) {
      e.printStackTrace();
      zerr(e.toString());
    }


    open();



    //
    // config environment
    //

    declareIO(2, 0);

    String[] ia = new String[] { 	
      "0: (messages)  open/close, oscdomain",
      "1: (OSC)  commands/requests",
    };

    setInletAssist(ia);
    createInfoOutlet(false);


    zms();
  }  // OSCSend()




  //------------------------------------------------------------------- -o--
  // messages
  //

  public void anything(String m, Atom[] args) 
  { 
    //if (0 < verbose) { zarglist("anything", m, args); }

    if (1 == getInlet())  { send(m, args); }
  }  // anything()



  // open()
  // 
  // Instantiate server object.
  //
  public void open() 
  {
    zsm("open");


    // server/listener
    //
    try {
      if (null == server) {
        server = OSCServer.newUsing(OSCServer.UDP);

	server.start();
	zi("target " + send_targetId +" @ "+ send_ip +":"+ send_port +" ("+ send_prefix +")");
      }


    } catch (Exception e) {
      e.printStackTrace();
      zerr(e.toString());
    }
      

    zms();
  }  // open();



  public void close() 
  {
    zsm("close");

    try {
      if (null != server) {
	server.stop();

	server.dispose();
	server = null;

	zi("target " + send_targetId +" @ "+ send_ip +":"+ send_port +" ("+ send_prefix +")");
      }

    } catch (Exception e) {
      e.printStackTrace();
      zerr(e.toString());
    }

    zms();
  }  // close()



  public void oscdomain(String prefix)  { setoscdomain(prefix); }




  //------------------------------------------------------------------- -o--
  // helper methods
  //

  private void setoscdomain(String prefix)
  {
    if (TargetType.CLIENT == send_targettype) {
      send_prefix = "/" + prefix +"/C/"+ send_targetId + "/";
    } else {
      send_prefix = "/" + prefix +"/S/";
    }
  }



  // send()
  //
  // Max message format: <AEcommand/AErequest> <arg(s)>
  // AE commands/requests that indicate subsystems MUST be separated by
  //   slashes ('/').
  //
  private void send(String aemsg, Atom[] args)
  {
    zsm("send");

    if (aemsg.equals("list")) {
      zerr("Outgoing message lacks request token.  Skipping..."); 
      return;
    }


    String request = send_prefix + aemsg;
    String arglist = Atom.toOneString(args);
    Object[] msgdata;


    try {
      //zd("SIZE="+args.length);
      //zd("AEMSG/ATOMS :: " + aemsg +" / "+ arglist);

      if (0 < verbose) { 
        zi(request +" "+ arglist 
	      +"  -->  "+ send_targetId 
	      +" @ "+ send_addr.getHostAddress() +":"+ send_port); 
      }


      if (0 == args.length) {
        msgdata = new Object[] { };
      } else {
        msgdata = new Object[] { arglist };
      }

      server.send( 
        new OSCMessage(request, msgdata), 
	send_sockaddr
      );


    } catch (Exception e) {
      e.printStackTrace();
      zerr(e.toString());
    }

    zms();
  }  // send()



  //------------------------------ -o-
  // local alternative to DebugLabel_M().  z.<method> --> z<method>
  //

  private String  zclass = "OSCSend",
  		  zfuncname = null;

  private void zsm(String s)  { zfuncname = s; }
  private void zms()  	      { zfuncname = null; }

  private void zpr(String label, String msg, boolean isErr, boolean isFatal)  
  { 
    String s = label +"\t\t\t\t\t"+ zclass +"."+ zfuncname +" :: "+ msg;
    if (isErr)  { error(s); if (isFatal) {bail("");} }
    else        { post(s); }
  }

  private void zd(String msg)  { zpr("DEBUG", msg, false, false); }
  private void zi(String msg)  { zpr("INFO", msg, false, false); }
  private void zerr(String msg)  { zpr("ERR", msg, true, false); }
  private void zferr(String msg)  { zpr("ERR", msg, true, true); }

  private void zarglist(Atom[] a) {
    zi("ARGS(" + a.length + ")\t\t" + Atom.toOneString(a)); 
  }

}  // public class OSCSend extends MaxObject
  
